program es_scalapack
use mpi
implicit none

integer :: i, j
! MPI
integer :: ierr, nprocs, myrank
! BLACS/SCALAPACK
integer :: context, nprow, npcol, iam, myrow, mycol, info
external :: blacs_exit, blacs_gridexit, blacs_gridinfo, descinit, numroc
external :: pdelset, pdelget, pdlawrite, pdlaread
integer :: numroc
! Matrix
integer :: M, N, MB, NB, Nloc, Mloc, aval
parameter ( M=10, N=10, MB=2, NB=2 )
real*8 , allocatable :: A(:,:), B(:), work(:)
integer, allocatable :: ipiv(:)
integer :: descA(9), descB(9)
real*8 :: t1, t2, dt
!character(len=25) :: filein, fileout

! Initialize MPI environment
call MPI_Init(ierr)
call MPI_Comm_Size(mpi_comm_world,nprocs,ierr)
call MPI_Comm_Rank(mpi_comm_world,myrank,ierr)

! Initialize a BLACS context
call sl_init_rm( iam, context, nprow, npcol )
call blacs_gridinfo( context, nprow, npcol, myrow, mycol )

! Computation of local matrix size
Mloc = numroc( M, MB, myrow, 0, nprow )
Nloc = numroc( N, NB, mycol, 0, npcol )
allocate( A( Mloc, Nloc ) )
allocate( B( Mloc ) )
allocate( ipiv( Mloc+MB ) )
allocate( work( M ) )

call descinit(descA, M, N, MB, NB, 0, 0, context, max(1,Mloc), info)
call descinit(descB, M, 1, MB,  1, 0, 0, context, max(1,Mloc), info)

! Vector and matrix elements generation
do j=1,N
  do i=1,M
    if(i.eq.j) then
      call pdelset( A, i, j, descA, dble(10000) )
   else
      call pdelset( A, i, j, descA, (dble(i)+dble(j)/dble(2.0)) )
   endif
  enddo
enddo
do i=1,M
  call pdelset( B, i, 1, descB, dble(207-i) )
enddo

!call pslaread('matrix.dat', A, descA, 0, 0, work)
call pdlawrite('A.dat', M, N, A, 1, 1, descA, 0, 0, work)
call pdlawrite('b.dat', M, 1, B, 1, 1, descB, 0, 0, work)

!!! for benchmarking !!!
call MPI_Barrier( mpi_comm_world, ierr)
t1 = MPI_wtime()
!!!!!!!!!!!!!!!!!!!!!!!!

call pdgesv( N, 1, A, 1, 1, descA, ipiv, B, 1, 1, descB, info )

!!! for benchmarking !!!
call MPI_Barrier( mpi_comm_world, ierr)
t2 = MPI_wtime()
dt = t2-t1
if(myrank.eq.0) write(*,*) nprocs, dt
!!!!!!!!!!!!!!!!!!!!!!!!

call pdlawrite('Af.dat', M, N, A, 1, 1, descA, 0, 0, work)
call pdlawrite('x.dat', M, 1, B, 1, 1, descB, 0, 0, work)

deallocate( A )
deallocate( B )
deallocate( ipiv )
deallocate( work )

! Close BLACS environment
call blacs_gridexit( context )
call blacs_exit( 0 )

! Close MPI environment
!call MPI_Finalize(ierr)

end program es_scalapack

!  file: SL_INIT_RM.f
      SUBROUTINE SL_INIT_RM( IAM, ICTXT, NPROW, NPCOL )
!                            OUT   OUT    OUT    OUT
!
!     .. Scalar Arguments ..
      INTEGER IAM, ICTXT, NPCOL, NPROW
!     ..
!
!  Purpose
!  =======
!
!  SL_INIT_RM initializes an NPROW x NPCOL process grid using a row-major
!  ordering (RM)  of  the  processes. This routine retrieves a default system
!  context which will include all available processes.
!
!  Arguments
!  =========
!
!  IAM     (global output) INTEGER
!          IAM is the rank of the procces
!
!  ICTXT   (global output) INTEGER
!          ICTXT specifies the BLACS context handle identifying the
!          created process grid.  The context itself is global.
!
!  NPROW   (global input) INTEGER
!          NPROW specifies the number of process rows in the grid
!          that will be created.  
!
!  NPCOL   (global input) INTEGER
!          NPCOL specifies the number of process columns in the grid
!          that will be created.
!
!  =====================================================================
!
!     .. Local Scalars ..
      INTEGER NPROC
!     ..
!     .. External Subroutines ..
      EXTERNAL BLACS_GET, BLACS_GRIDINIT, BLACS_PINFO
!     ..
!     .. Executable Statements ..
!
!     Get starting information
!
      CALL BLACS_PINFO( IAM, NPROC )
!                       OUT   OUT
!
! -----    Set the dimension of the 2d processors grid.
!
      call gridsetup(NPROC,NPROW,NPCOL)
!                      IN   OUT   OUT
!
!     Define process grid
!
      CALL BLACS_GET( -1, 0, ICTXT )
!                     IN IN   OUT
      CALL BLACS_GRIDINIT( ICTXT, 'Row-major', NPROW, NPCOL )
!                            IN     IN           IN     IN
!
      RETURN
!
!     End of SL_INIT_RM
!
      END

!-----------------------------------------------------------------------
!
      subroutine gridsetup(nproc,nprow,npcol)
!
! This subroutine factorizes the number of processors (nproc)
! into nprow and npcol, that are the sizes of the 2d processors mesh.
!
      implicit none
      integer nproc,nprow,npcol
      integer sqrtnp,i 

      sqrtnp = int( sqrt( dble(nproc) ) + 1 )
      do i=1,sqrtnp
        if(mod(nproc,i).eq.0) nprow = i
      end do
      npcol = nproc/nprow

      return
      end

